@@ -49,6 +49,8 @@ from nuitka.Progress import (
 from nuitka.PythonVersions import (
     getPythonABI,
     getSupportedPythonVersions,
+    getSystemPrefixPath,
+    getSystemStaticLibPythonPath,
     python_version,
     python_version_str,
 )
@@ -64,6 +66,7 @@ from nuitka.utils.FileOperations import (
 )
 from nuitka.utils.Importing import getSharedLibrarySuffix
 from nuitka.utils.ModuleNames import ModuleName
+from nuitka.Version import getCommercialVersion, getNuitkaVersion
 
 from . import ModuleRegistry, Options, OutputDirectories, TreeXML
 from .build import SconsInterface
@@ -84,6 +87,8 @@ def _createNodeTree(filename):
 
     """
 
+    # Many cases to deal with, pylint: disable=too-many-branches
+
     # First, build the raw node tree from the source code.
     main_module = Building.buildModuleTree(
         filename=filename,
@@ -106,7 +111,15 @@ def _createNodeTree(filename):
         removeDirectory(path=standalone_dir, ignore_errors=True)
         makePath(standalone_dir)
 
-    deleteFile(path=OutputDirectories.getResultFullpath(), must_exist=False)
+    # Delete result file, to avoid confusion with previous build and to
+    # avoid locking issues after the build.
+    deleteFile(
+        path=OutputDirectories.getResultFullpath(onefile=False), must_exist=False
+    )
+    if Options.isOnefileMode():
+        deleteFile(
+            path=OutputDirectories.getResultFullpath(onefile=True), must_exist=False
+        )
 
     # Second, do it for the directories given.
     for plugin_filename in Options.getShallFollowExtra():
@@ -292,7 +305,10 @@ def makeSourceDirectory():
             if module.getFullName() == any_case_module:
                 break
         else:
-            general.warning("Not recursing to unused '%s'." % any_case_module)
+            general.warning(
+                "Did not follow import to unused '%s', consider include options."
+                % any_case_module
+            )
 
     # Prepare code generation, i.e. execute finalization for it.
     for module in ModuleRegistry.getDoneModules():
@@ -381,21 +397,11 @@ def makeSourceDirectory():
         source_code=constants_body_code,
     )
 
-    for filename, source_code in Plugins.getExtraCodeFiles().items():
-        target_dir = os.path.join(source_dir, "plugins")
-
-        if not os.path.isdir(target_dir):
-            makePath(target_dir)
-
-        writeSourceCode(
-            filename=os.path.join(target_dir, filename), source_code=source_code
-        )
-
 
 def runSconsBackend(quiet):
     # Scons gets transported many details, that we express as variables, and
     # have checks for them, leading to many branches and statements,
-    # pylint: disable=too-many-branches
+    # pylint: disable=too-many-branches,too-many-statements
 
     asBoolStr = SconsInterface.asBoolStr
 
@@ -411,7 +417,7 @@ def runSconsBackend(quiet):
         "trace_mode": asBoolStr(Options.shallTraceExecution()),
         "python_version": python_version_str,
         "target_arch": Utils.getArchitecture(),
-        "python_prefix": getDirectoryRealPath(sys.prefix),
+        "python_prefix": getDirectoryRealPath(getSystemPrefixPath()),
         "nuitka_src": SconsInterface.getSconsDataPath(),
         "module_count": "%d"
         % (
@@ -422,10 +428,10 @@ def runSconsBackend(quiet):
     }
 
     if not Options.shallMakeModule():
-        options["result_exe"] = OutputDirectories.getResultFullpath()
+        options["result_exe"] = OutputDirectories.getResultFullpath(onefile=False)
 
-    if Options.shallUseStaticLibPython():
-        options["static_libpython"] = asBoolStr(True)
+    if Options.shallUseStaticLibPython() and getSystemStaticLibPythonPath() is not None:
+        options["static_libpython"] = getSystemStaticLibPythonPath()
 
     if Options.isStandaloneMode():
         options["standalone_mode"] = asBoolStr(True)
@@ -433,6 +439,15 @@ def runSconsBackend(quiet):
     if Options.isOnefileMode():
         options["onefile_mode"] = asBoolStr(True)
 
+    if Options.isWindowsOnefileTempDirMode():
+        options["onefile_temp_mode"] = asBoolStr(True)
+
+    if Options.getForcedStdoutPath():
+        options["forced_stdout_path"] = Options.getForcedStdoutPath()
+
+    if Options.getForcedStderrPath():
+        options["forced_stderr_path"] = Options.getForcedStderrPath()
+
     if Options.shallTreatUninstalledPython():
         options["uninstalled_python"] = asBoolStr(True)
 
@@ -464,7 +479,7 @@ def runSconsBackend(quiet):
     if sys.flags.bytes_warning:
         options["python_sysflag_bytes_warning"] = asBoolStr(True)
 
-    if int(os.environ.get("NUITKA_SITE_FLAG", "no_site" in Options.getPythonFlags())):
+    if int(os.environ.get("NUITKA_SITE_FLAG", Options.hasPythonFlagNoSite())):
         options["python_sysflag_no_site"] = asBoolStr(True)
 
     if "trace_imports" in Options.getPythonFlags():
@@ -612,9 +627,18 @@ def compileTree():
 
     general.info("Running data composer tool for optimal constant value handling.")
 
-    # TODO: On Windows, we could run this in parallel to Scons, on Linux we need it
-    # for linking.
-    runDataComposer(source_dir)
+    blob_filename = runDataComposer(source_dir)
+    Plugins.onDataComposerResult(blob_filename)
+
+    for filename, source_code in Plugins.getExtraCodeFiles().items():
+        target_dir = os.path.join(source_dir, "plugins")
+
+        if not os.path.isdir(target_dir):
+            makePath(target_dir)
+
+        writeSourceCode(
+            filename=os.path.join(target_dir, filename), source_code=source_code
+        )
 
     general.info("Running C level backend compilation via Scons.")
 
@@ -664,7 +688,10 @@ def main():
     # Main has to fulfill many options, leading to many branches and statements
     # to deal with them.  pylint: disable=too-many-branches,too-many-statements
     if not Options.shallDumpBuiltTreeXML():
-        general.info("Starting Python compilation.")
+        general.info(
+            "Starting Python compilation with Nuitka %r on Python %r commercial %r."
+            % (getNuitkaVersion(), python_version_str, getCommercialVersion())
+        )
 
     filename = Options.getPositionalArgs()[0]
 
