@@ -35,6 +35,7 @@ from nuitka.PythonVersions import (
 from nuitka.Tracing import postprocessing_logger
 from nuitka.utils.FileOperations import (
     getFileContents,
+    makePath,
     removeFileExecutablePermission,
 )
 from nuitka.utils.SharedLibraries import callInstallNameTool
@@ -74,6 +75,9 @@ class IconDirectoryEntry(ctypes.Structure):
 
 
 class IconGroupDirectoryEntry(ctypes.Structure):
+    # Make sure the don't have padding issues.
+    _pack_ = 2
+
     _fields_ = (
         ("width", ctypes.c_char),
         ("height", ctypes.c_char),
@@ -87,7 +91,7 @@ class IconGroupDirectoryEntry(ctypes.Structure):
 
 
 def readFromFile(readable, c_struct):
-    """ Read ctypes structures from input. """
+    """Read ctypes structures from input."""
 
     result = c_struct()
     chunk = readable.read(ctypes.sizeof(result))
@@ -95,12 +99,14 @@ def readFromFile(readable, c_struct):
     return result
 
 
-def addWindowsIconFromIcons():
+def _addWindowsIconFromIcons(onefile):
+    # Relatively detailed handling, pylint: disable=too-many-locals
+
     icon_group = 1
     image_id = 1
     images = []
 
-    result_filename = OutputDirectories.getResultFullpath()
+    result_filename = OutputDirectories.getResultFullpath(onefile=onefile)
 
     for icon_spec in Options.getIconPaths():
         if "#" in icon_spec:
@@ -110,6 +116,48 @@ def addWindowsIconFromIcons():
             icon_path = icon_spec
             icon_index = None
 
+        icon_path = os.path.normcase(icon_path)
+
+        if not icon_path.endswith(".ico"):
+            postprocessing_logger.info("Not in icon format, converting.")
+
+            if icon_index is not None:
+                postprocessing_logger.sysexit(
+                    "Cannot specify indexes with non-ico format files in '%s'."
+                    % icon_spec
+                )
+
+            try:
+                import imageio
+            except ImportError:
+                postprocessing_logger.sysexit(
+                    "Need to install imageio to use non-ico icon file in '%s'."
+                    % icon_spec
+                )
+
+            try:
+                image = imageio.imread(icon_path)
+            except ValueError:
+                postprocessing_logger.sysexit(
+                    "Unsupported file format for imageio in '%s', use e.g. PNG files."
+                    % icon_spec
+                )
+
+            icon_build_path = os.path.join(
+                OutputDirectories.getSourceDirectoryPath(onefile=onefile),
+                "icons",
+            )
+
+            makePath(icon_build_path)
+
+            converted_icon_path = os.path.join(
+                icon_build_path,
+                "icon-%d.ico" % image_id,
+            )
+
+            imageio.imwrite(converted_icon_path, image)
+            icon_path = converted_icon_path
+
         with open(icon_path, "rb") as icon_file:
             # Read header and icon entries.
             header = readFromFile(icon_file, IconDirectoryHeader)
@@ -121,14 +169,14 @@ def addWindowsIconFromIcons():
             if icon_index is not None:
                 if icon_index > len(icons):
                     postprocessing_logger.sysexit(
-                        "Error, referenced icon index %d in file %r with only %d icons."
+                        "Error, referenced icon index %d in file '%s' with only %d icons."
                         % (icon_index, icon_path, len(icons))
                     )
 
                 icons[:] = icons[icon_index : icon_index + 1]
 
             postprocessing_logger.info(
-                "Adding %d icon(s) from icon file %r." % (len(icons), icon_spec)
+                "Adding %d icon(s) from icon file '%s'." % (len(icons), icon_spec)
             )
 
             # Image data are to be scanned from places specified icon entries
@@ -179,10 +227,76 @@ def addWindowsIconFromIcons():
 version_resources = {}
 
 
-def executePostProcessing():
-    # These is a bunch of stuff to consider, pylint: disable=too-many-branches
+def executePostProcessingResources(manifest, onefile):
+    """Adding Windows resources to the binary.
 
-    result_filename = OutputDirectories.getResultFullpath()
+    Used for both onefile and not onefile binary, potentially two times.
+    """
+    result_filename = OutputDirectories.getResultFullpath(onefile=onefile)
+
+    # TODO: Maybe make these different for onefile and not onefile.
+    if (
+        Options.shallAskForWindowsAdminRights()
+        or Options.shallAskForWindowsUIAccessRights()
+    ):
+        if manifest is None:
+            manifest = getDefaultWindowsExecutableManifest()
+
+        if Options.shallAskForWindowsAdminRights():
+            manifest.addUacAdmin()
+
+        if Options.shallAskForWindowsUIAccessRights():
+            manifest.addUacUiAccess()
+
+    if manifest is not None:
+        manifest.addResourceToFile(result_filename, logger=postprocessing_logger)
+
+    if (
+        Options.getWindowsVersionInfoStrings()
+        or Options.getWindowsProductVersion()
+        or Options.getWindowsFileVersion()
+    ):
+        version_resources.update(
+            addVersionInfoResource(
+                string_values=Options.getWindowsVersionInfoStrings(),
+                product_version=Options.getWindowsProductVersion(),
+                file_version=Options.getWindowsFileVersion(),
+                file_date=(0, 0),
+                is_exe=not Options.shallMakeModule(),
+                result_filename=result_filename,
+                logger=postprocessing_logger,
+            )
+        )
+
+    # Attach icons from template file if given.
+    template_exe = Options.getWindowsIconExecutablePath()
+    if template_exe is not None:
+        res_copied = copyResourcesFromFileToFile(
+            template_exe,
+            target_filename=result_filename,
+            resource_kinds=(RT_ICON, RT_GROUP_ICON),
+        )
+
+        if res_copied == 0:
+            postprocessing_logger.warning(
+                "The specified icon template executable %r didn't contain anything to copy."
+                % template_exe
+            )
+        else:
+            postprocessing_logger.warning(
+                "Copied %d icon resources from %r." % (res_copied, template_exe)
+            )
+    else:
+        _addWindowsIconFromIcons(onefile=onefile)
+
+
+def executePostProcessing():
+    """Postprocessing of the resulting binary.
+
+    These are in part required steps, not usable after failure.
+    """
+
+    result_filename = OutputDirectories.getResultFullpath(onefile=False)
 
     if not os.path.exists(result_filename):
         postprocessing_logger.sysexit(
@@ -191,54 +305,15 @@ def executePostProcessing():
 
     if isWin32Windows():
         if not Options.shallMakeModule():
-            needs_manifest = False
-            manifest = None
-
             if python_version < 0x300:
                 # Copy the Windows manifest from the CPython binary to the created
                 # executable, so it finds "MSCRT.DLL". This is needed for Python2
                 # only, for Python3 newer MSVC doesn't hide the C runtime.
                 manifest = getWindowsExecutableManifest(sys.executable)
+            else:
+                manifest = None
 
-                if manifest is not None:
-                    needs_manifest = True
-
-            if (
-                Options.shallAskForWindowsAdminRights()
-                or Options.shallAskForWindowsUIAccessRights()
-            ):
-                needs_manifest = True
-
-                if manifest is None:
-                    manifest = getDefaultWindowsExecutableManifest()
-
-                if Options.shallAskForWindowsAdminRights():
-                    manifest.addUacAdmin()
-
-                if Options.shallAskForWindowsUIAccessRights():
-                    manifest.addUacUiAccess()
-
-            if needs_manifest:
-                manifest.addResourceToFile(
-                    result_filename, logger=postprocessing_logger
-                )
-
-        if (
-            Options.getWindowsVersionInfoStrings()
-            or Options.getWindowsProductVersion()
-            or Options.getWindowsFileVersion()
-        ):
-            version_resources.update(
-                addVersionInfoResource(
-                    string_values=Options.getWindowsVersionInfoStrings(),
-                    product_version=Options.getWindowsProductVersion(),
-                    file_version=Options.getWindowsFileVersion(),
-                    file_date=(0, 0),
-                    is_exe=not Options.shallMakeModule(),
-                    result_filename=result_filename,
-                    logger=postprocessing_logger,
-                )
-            )
+            executePostProcessingResources(manifest=manifest, onefile=False)
 
         source_dir = OutputDirectories.getSourceDirectoryPath()
 
@@ -252,27 +327,6 @@ def executePostProcessing():
             logger=postprocessing_logger,
         )
 
-        # Attach icons from template file if given.
-        template_exe = Options.getWindowsIconExecutablePath()
-        if template_exe is not None:
-            res_copied = copyResourcesFromFileToFile(
-                template_exe,
-                target_filename=result_filename,
-                resource_kinds=(RT_ICON, RT_GROUP_ICON),
-            )
-
-            if res_copied == 0:
-                postprocessing_logger.warning(
-                    "The specified icon template executable %r didn't contain anything to copy."
-                    % template_exe
-                )
-            else:
-                postprocessing_logger.warning(
-                    "Copied %d icon resources from %r." % (res_copied, template_exe)
-                )
-        else:
-            addWindowsIconFromIcons()
-
     # On macOS, we update the executable path for searching the "libpython"
     # library.
     if (
