@@ -56,6 +56,7 @@ from nuitka.utils.AppDirs import getCacheDir
 from nuitka.utils.Execution import getNullInput, withEnvironmentPathAdded
 from nuitka.utils.FileOperations import (
     areSamePaths,
+    copyTree,
     getDirectoryRealPath,
     getFileContentByLine,
     getFileContents,
@@ -85,7 +86,7 @@ from nuitka.utils.ThreadedExecutor import ThreadPoolExecutor, waitWorkers
 from nuitka.utils.Timing import TimerReport
 
 from .DependsExe import detectDLLsWithDependencyWalker
-from .IncludedDataFiles import IncludedDataFile
+from .IncludedDataFiles import IncludedDataFile, makeIncludedDataFile
 
 
 def loadCodeObjectData(precompiled_filename):
@@ -584,7 +585,7 @@ _detected_python_rpath = None
 ldd_result_cache = {}
 
 
-def _detectBinaryPathDLLsPosix(dll_filename):
+def _detectBinaryPathDLLsPosix(dll_filename, package_name, original_dir):
     # This is complex, as it also includes the caching mechanism
     # pylint: disable=too-many-branches
 
@@ -607,7 +608,16 @@ def _detectBinaryPathDLLsPosix(dll_filename):
                 "$ORIGIN", os.path.dirname(sys.executable)
             )
 
-    with withEnvironmentPathAdded("LD_LIBRARY_PATH", _detected_python_rpath):
+    ld_library_path = OrderedSet()
+    if _detected_python_rpath:
+        ld_library_path.add(_detected_python_rpath)
+    ld_library_path.update(getPackageSpecificDLLDirectories(package_name))
+
+    if original_dir is not None:
+        ld_library_path.add(original_dir)
+        # ld_library_path.update(getSubDirectories(original_dir, ignore_dirs=("__pycache__",)))
+
+    with withEnvironmentPathAdded("LD_LIBRARY_PATH", *ld_library_path):
         process = subprocess.Popen(
             args=["ldd", dll_filename],
             stdin=getNullInput(),
@@ -707,7 +717,13 @@ def _detectBinaryPathDLLsPosix(dll_filename):
     sub_result = set(result)
 
     for sub_dll_filename in result:
-        sub_result = sub_result.union(_detectBinaryPathDLLsPosix(sub_dll_filename))
+        sub_result = sub_result.union(
+            _detectBinaryPathDLLsPosix(
+                dll_filename=sub_dll_filename,
+                package_name=package_name,
+                original_dir=original_dir,
+            )
+        )
 
     return sub_result
 
@@ -837,6 +853,25 @@ def _getCacheFilename(
 _scan_dir_cache = {}
 
 
+def getPackageSpecificDLLDirectories(package_name):
+    scan_dirs = OrderedSet()
+
+    if package_name is not None:
+        from nuitka.importing.Importing import findModule
+
+        package_dir = findModule(None, package_name, None, 0, False)[1]
+
+        if os.path.isdir(package_dir):
+            scan_dirs.add(package_dir)
+            scan_dirs.update(
+                getSubDirectories(package_dir, ignore_dirs=("__pycache__",))
+            )
+
+        scan_dirs.update(Plugins.getModuleSpecificDllPaths(package_name))
+
+    return scan_dirs
+
+
 def getScanDirectories(package_name, original_dir):
     # Many cases, pylint: disable=too-many-branches
 
@@ -971,7 +1006,11 @@ def detectBinaryDLLs(
         Utils.getOS() in ("Linux", "NetBSD", "FreeBSD", "OpenBSD")
         or Utils.isPosixWindows()
     ):
-        return _detectBinaryPathDLLsPosix(dll_filename=original_filename)
+        return _detectBinaryPathDLLsPosix(
+            dll_filename=original_filename,
+            package_name=package_name,
+            original_dir=os.path.dirname(original_filename),
+        )
     elif Utils.isWin32Windows():
         with TimerReport(
             message="Running depends.exe for %s took %%.2f seconds" % binary_filename,
@@ -1062,7 +1101,7 @@ def detectUsedDLLs(source_dir, standalone_entry_points, use_cache, update_cache)
 
 
 def fixupBinaryDLLPathsMacOS(binary_filename, dll_map, original_location):
-    """ For macOS, the binary needs to be told to use relative DLL paths """
+    """For macOS, the binary needs to be told to use relative DLL paths"""
 
     # There may be nothing to do, in case there are no DLLs.
     if not dll_map:
@@ -1286,7 +1325,7 @@ def _handleDataFile(dist_dir, tracer, included_datafile):
 
             for sub_dir in included_datafile.dest_path:
                 makePath(os.path.join(dist_dir, sub_dir))
-        elif included_datafile.kind == "datafile":
+        elif included_datafile.kind == "data_file":
             dest_path = os.path.join(dist_dir, included_datafile.dest_path)
 
             tracer.info(
@@ -1299,6 +1338,20 @@ def _handleDataFile(dist_dir, tracer, included_datafile):
 
             makePath(os.path.dirname(dest_path))
             shutil.copyfile(included_datafile.source_path, dest_path)
+        elif included_datafile.kind == "data_dir":
+            dest_path = os.path.join(dist_dir, included_datafile.dest_path)
+            makePath(os.path.dirname(dest_path))
+
+            copied = copyTree(included_datafile.source_path, dest_path)
+
+            tracer.info(
+                "Included data dir %r with %d files due to %s."
+                % (
+                    included_datafile.dest_path,
+                    len(copied),
+                    included_datafile.reason,
+                )
+            )
         else:
             assert False, included_datafile
     else:
@@ -1417,6 +1470,3 @@ def copyDataFiles(dist_dir):
                         )
 
                 # assert False, (module.getCompileTimeDirectory(), pkg_files)
-
-
-from .IncludedDataFiles import makeIncludedDataFile
