@@ -1,4 +1,4 @@
-#     Copyright 2015, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -32,61 +32,53 @@ to grow both.
 """
 
 
-
 import re
 
+from nuitka.__past__ import intern  # pylint: disable=I0021,redefined-builtin
 from nuitka.plugins.PluginBase import NuitkaPluginBase
 
 
 class NuitkaPluginPylintEclipseAnnotations(NuitkaPluginBase):
     plugin_name = "pylint-warnings"
+    plugin_desc = "Support PyLint / PyDev linting source markers"
 
     def __init__(self):
         self.line_annotations = {}
 
-    def onModuleSourceCode(self, module_name, source_code):
+    def checkModuleSourceCode(self, module_name, source_code):
         annotations = {}
 
-        for count, line in enumerate(source_code.split('\n')):
-            match = re.search(r"#.*pylint:\s*disable=\s*([\w,]+)", line)
+        for count, line in enumerate(source_code.split("\n"), 1):
+            match = re.search(r"#.*pylint:\s*disable=\s*([\w,-]+)", line)
 
             if match:
-                comment_only = line[:line.find('#')-1].strip() == ""
+                comment_only = line[: line.find("#") - 1].strip() == ""
 
                 if comment_only:
                     # TODO: Parse block wide annotations too.
                     pass
                 else:
-                    annotations[count+1] = set(
-                        match.strip()
-                        for match in
-                        match.group(1).split(',')
+                    annotations[count] = set(
+                        intern(match.strip()) for match in match.group(1).split(",")
                     )
 
         # Only remember them if there were any.
         if annotations:
             self.line_annotations[module_name] = annotations
 
-        # Do nothing to it.
-        return source_code
-
     def suppressUnknownImportWarning(self, importing, module_name, source_ref):
         annotations = self.line_annotations.get(importing.getFullName(), {})
 
         line_annotations = annotations.get(source_ref.getLineNumber(), ())
 
-        if "F0401" in line_annotations:
+        if "F0401" in line_annotations or "import-error" in line_annotations:
             return True
 
         return False
 
 
 class NuitkaPluginDetectorPylintEclipseAnnotations(NuitkaPluginBase):
-    plugin_name = "pylint-warnings"
-
-    @staticmethod
-    def isRelevant():
-        return True
+    detector_for = NuitkaPluginPylintEclipseAnnotations
 
     def onModuleSourceCode(self, module_name, source_code):
         if re.search(r"#\s*pylint:\s*disable=\s*(\w+)", source_code):
